;;; -*- lisp -*-

(in-package :cl-webobjects)

;(print *this-file*)

(defgeneric lisplet-register-page-location (obj page-class location))
(defgeneric lisplet-register-resource-location (obj uri url))

(defgeneric lisplet-dispatch-request (obj))
(defgeneric lisplet-dispatch-method (obj))


(defclass lisplet ()
  ((base-path :initarg :base-path
	      :reader lisplet-base-path)
   (realm :initarg :realm
	  :reader lisplet-realm)   
   (pages :initform (make-hash-table :test 'equal)
	  :reader lisplet-pages)
   (page404 :initarg :page404
	    :accessor lisplet-page404))  
  (:default-initargs :realm nil :page404 (make-instance 'page404)))

(defun build-lisplet-location (lisplet location)
  (let ((server-base-path *woserver-base-path*)
	(base-path (lisplet-base-path lisplet)))
    (setf location (format nil "~a/~a" base-path location))
    (unless (null server-base-path)
      (setf location (format nil "~a~a" server-base-path location)))
    location))

(defmethod lisplet-register-page-location ((obj lisplet) page-class location)
  (let ((pages (lisplet-pages obj))
	(new-location (build-lisplet-location obj location)))
    (setf (gethash new-location pages) (create-prefix-dispatcher 
				    new-location
				    #'(lambda () 										
					(with-output-to-string 
					    (*standard-output*)
					  (page-render (make-instance page-class :lisplet obj :url new-location))))))))

(defmethod lisplet-register-resource-location ((obj lisplet) resource-path location)
  (let ((pages (lisplet-pages obj))
	(new-location (build-lisplet-location obj location)))
    (setf (gethash new-location pages) (create-folder-dispatcher-and-handler
				    new-location
				    resource-path))))


(defmethod lisplet-dispatch-request ((obj lisplet))
  (let ((pages (lisplet-pages obj)))
    (loop for dispatcher being the hash-value of pages
       for action = (funcall dispatcher *request*)
       when action return (progn
			    (setf (request-realm *request*) (lisplet-realm obj))			    
			    (funcall action)))))


(defmethod lisplet-dispatch-method ((obj lisplet))
  (let ((page404 (lisplet-page404 obj))
	(result nil))
    (progn 
      (setf result (lisplet-dispatch-request obj)) 
      (when (null result)
	(setf result (with-output-to-string (*standard-output*) (page-render page404))))
      result)))
