;;; -*- lisp -*-
;;; -- tags.lisp --

(in-package :cl-webobjects)



(defgeneric page-req-parameter (obj name &optional as-list)
  (:documentation "This method returns a request parameter given by NAME searching first 
into post parameters and, if no parameter found, into get prarmeters.
The optional function parameter AS-LIST if true returns the result as list.
When AS-LIST is true, if the searched parameter is found more then once, a list with 
all valuse given to param NAME is returned.
 - OBJ is the page instance that must be given.
 - NAME The parameter to search
 - AS-LIST If true the result is returned as list, if false as string. Default: false"))

(defgeneric page-json-id-list (obj)
  (:documentation "This internal method is called to get a list of all the components by their id, that must be updated when 
an xhr request is sent from the browser.
 - OBJ is the page instance that must be given"))

(defgeneric page-content (obj) 
  (:documentation "This method returns the page content to be redered.
 - OBJ is the page instance that must be given"))

(defgeneric page-init (obj) 
  (:documentation "Internal method for page initialization.
 - OBJ is the page instance that must be given"))

(defgeneric page-render (obj)
  (:documentation "This method is the main method fired from the framework to render the desired page and to handle all the request cycle.
 - OBJ is the page instance that must be given"))

(defgeneric page-init-injections (pobj)
  (:documentation "This internal method is called during the request cycle phase to reset page slots that 
must be reinitialized during sub-phases (rewinding, pre-rendering, rendering).
 - OBJ is the page instance that must be given"))

(defgeneric page-render-headings (obj)
  (:documentation "This internal method renders the html first lines that determine if the page is a html or a xhtml, along with the schema definition.
 - OBJ is the page instance that must be given"))

(defgeneric page-request-parameters (obj) 
  (:documentation "This internal method builds the get and post parameters into an hash table.
 - OBJ is the page instance that must be given"))

(defgeneric page-print-tabulation (obj)
  (:documentation "This internal method is called during the rendering phase if tabulation is enabled. It writes the right amount
of tabs chars to indent the page.
 - OBJ is the page instance that must be given"))

(defgeneric page-newline (obj)
  (:documentation "This internal method simply writes the rest of page content on a new line when needed.
 - OBJ is the page instance that must be given"))

(defgeneric page-format (obj str &rest rest)
  (:documentation "This internal method is the replacement of the FORMAT function. It is aware
of an xhr request when the reply must be given as a json object. It also uses the default page output stream 
to render the output. 
 - OBJ is the page instance that must be given
 - STR The format control
 - REST The format arguments
See http://www.lisp.org/HyperSpec/Body/fun_format.html#format for more info."))

(defgeneric page-format-raw (obj str &rest rest)
  (:documentation "This internal method is the replacement of the FORMAT.
The difference with PAGE-FORMAT is that it prints out the result ignoring the json directive. 
It also uses the default page output stream as PAGE-FORMAT does to render the output. 
 - OBJ is the page instance that must be given
 - STR The format control
 - REST The format arguments
See http://www.lisp.org/HyperSpec/Body/fun_format.html#format for more info."))

(defgeneric page-body-init-scripts (page-obj)
  (:documentation "During the render phase wocomponent instances inject their initialization scripts (javascript)
that will be evaluated when the page has been loaded. 
This internal method is called to render these scripts.
 - PAGE-OBJ is the page instance that must be given")) 

(defgeneric htbody-init-scripts-tag (page-obj)
  (:documentation "Encloses the init inscance scripts injected into the page into a <script> tag component
See PAGE-BODY-INIT-SCRIPTS form more info.
 - PAGE-OBJ is the page instance that must be given"))

(defgeneric htcomponent-rewind (obj page-obj)
  (:documentation "This internal method is the first called during the request cycle phase. 
It is evaluated when a form action or an action-link action is fired. It is used to update all visit objects slots.
 - OBJ is the htcomponent instance that must be rewound
 - PAGE-OBJ is the page instance that must be given"))

(defgeneric htcomponent-prerender (obj page-obj)
  (:documentation "This internal method is the second sub phase during the request cycle phase. 
It is used to inject all wocomponent class scripts and stylesheets into the owner page.
 - OBJ is the htcomponent instance that must be prerendered
 - PAGE-OBJ is the page instance that must be given"))

(defgeneric htcomponent-render (obj page-obj)
  (:documentation "This internal method is the last called during the request cycle phase. 
It is used to effectively render the component into the page.
 - OBJ is the htcomponent instance that must be rendered
 - PAGE-OBJ is the page instance that must be given"))

(defgeneric htcomponent-can-print (obj)
  (:documentation "This internal method is used in an xhr call to determine 
if a component may be rendered into the reply
 - OBJ is the htcomponent instance"))

(defgeneric htcomponent-json-print-start-component (obj)
  (:documentation "Internal method called to render the json reply during the render cycle phase
on component start.
 - OBJ is the htcomponent instance"))

(defgeneric htcomponent-json-print-end-component (obj)
    (:documentation "Internal method called to render the json reply during the render cycle phase
on component end.
 - OBJ is the htcomponent instance"))

(defgeneric tag-render-starttag (obj page-obj) 
  (:documentation "Internal method to print out the opening html tag during the render phase
 - OBJ is the tag instance
 - PAGE-OBJ the page instance"))

(defgeneric tag-render-endtag (obj page-obj)
  (:documentation "Internal method to print out the closing html tag during the render phase
 - OBJ is the tag instance
 - PAGE-OBJ the page instance"))

(defgeneric tag-render-attributes (obj page-obj)
  (:documentation "Internal method to print out the attributes of an html tag during the render phase
 - OBJ is the tag instance
 - PAGE-OBJ the page instance"))

(defgeneric (setf htcomponent-page) (page-obj obj)
  (:documentation "Internal method to set the component owner page and to assign 
an unique id attribute when provided.
 - OBJ is the tag instance
 - PAGE-OBJ the page instance"))

(defgeneric wocomponent-parameter-value (obj key)
  (:documentation "Returns the value of a parameter passed to the wocomponent initialization
function (the one generated with DEFCOMPONENT) or :UNDEFINED if not passed.
 - OBJ is the wocomponent instance
 - KEY the parameter key to query"))

(defgeneric wocomponent-check-parameters(obj)
  (:documentation "This internal method check if all :REQUIRED parameters are provided
 - OBJ is the wocomponent instance"))

(defgeneric wocomponent-parameters(obj)
  (:documentation "This method returns class formal parameters as an alist (formal parameters are the ones expected by the component)
 - OBJ is the wocomponent instance"))
(defgeneric wocomponent-informal-parameters(obj)
  (:documentation "This method returns class informal parameters as an alist (informal parameters are the ones not expected by the component, 
usually rendered as tag attributes withot any kind of evaluation)
 - OBJ is the wocomponent instance"))

(defgeneric wocomponent-before-rewind (obj page-obj)
  (:documentation "Method called by the framework before the rewinding phase. It is intended to be eventually overridden in descendant classes.
 - OBJ is the tag instance
 - PAGE-OBJ the page instance"))

(defgeneric wocomponent-after-rewind (obj page-obj)
  (:documentation "Method called by the framework after the rewinding phase. It is intended to be eventually overridden in descendant classes.
 - OBJ is the tag instance
 - PAGE-OBJ the page instance"))
(defgeneric wocomponent-before-prerender (obj page-obj)
  (:documentation "Method called by the framework before the pre-rendering phase. It is intended to be eventually overridden in descendant classes.
 - OBJ is the tag instance
 - PAGE-OBJ the page instance"))

(defgeneric wocomponent-after-prerender (obj page-obj)
  (:documentation "Method called by the framework after the pre-rendering phase. It is intended to be eventually overridden in descendant classes.
 - OBJ is the tag instance
 - PAGE-OBJ the page instance"))
(defgeneric wocomponent-before-render (obj page-obj)
  (:documentation "Method called by the framework before the rendering phase. It is intended to be eventually overridden in descendant classes.
 - OBJ is the tag instance
 - PAGE-OBJ the page instance"))

(defgeneric wocomponent-after-render (obj page-obj)
  (:documentation "Method called by the framework after the rendering phase. It is intended to be eventually overridden in descendant classes.
 - OBJ is the tag instance
 - PAGE-OBJ the page instance"))

(defvar *woserver-base-path* nil)

(defvar *html-4.01-strict* "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">" "Page doctype as HTML 4.01 STRICT")

(defvar *html-4.01-transitional* "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">" "Page doctype as HTML 4.01 TRANSITIONAL")

(defvar *html-4.01-frameset* "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Frameset//EN\" \"http://www.w3.org/TR/html4/frameset.dtd\">" "Page doctype as HTML 4.01 FRAMESET")

(defvar *xhtml-1.0-strict* "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">" "Page doctype as HTML 4.01 XHTML")

(defvar *xhtml-1.0-transitional* "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">" "Page doctype as XHTML 4.01 TRANSITIONAL")

(defvar *xhtml-1.0-frameset* "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Frameset//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd\">" "Page doctype as XHTML 4.01 FRAMESET")

(defvar *default-encoding* "UTF-8" "Page default encoding (if no changes 'UTF-8')")

(defvar *rewind-parameter* "rewindobject" "The request parameter for the object asking for a rewind action")

(defvar *empty-tags*
  (list "area" "base" "basefont" "br" "col" "frame"
	"hr" "img" "input" "isindex" "meta"
	"param" "link"))

(defun request-id-table-map ()
  "Holds an hash table of used components/tags id as keys and the number of their occurrences as values. 
So if you have a :id \"compId\" given to a previous component, the second 
time this id will be used, it will be rendered as \"compId1\", the third time will be \"compId2\" and so on"
  (when (boundp '*request*)    
    (let ((id-table-map (aux-request-value :id-table-map)))
      (if (null id-table-map)
	  (progn 
	    (setf (aux-request-value :id-table-map) (make-hash-table :test 'equal)))
	  id-table-map))))
    
(defun strings-to-jsarray (strings)
  "Transforms a list of strings into a javascript array."
  (let ((st-size (length strings))
	(items ""))
    (cond ((= st-size 0) "[]")
	  ((= st-size 1) (format nil "[~a]" (prin1-to-string (first strings))))
	  (t (format nil (format nil "[~a~a]" 
				 (prin1-to-string (first strings))
				 (progn
				   (dolist (str (rest strings))
				     (setf items (format nil "~a,~a"
							 items (prin1-to-string str))))
				   items)))))))
				 

(defun reset-request-id-table-map ()
  "This function resets the ID-TABLE-MAP built during the request cycle to handle id uniqueness.
See REQUEST-ID-TABLE-MAP for more info."
  (when (boundp '*request*)    
    (setf (aux-request-value :id-table-map) (make-hash-table :test 'equal))))


(defun parse-htcomponent-function (function-body)
  "This function parses attributes passed to a htcomponent creation function"
  (let ((attributes)
	(body))
    (loop for last-elem = nil then elem
       for elem in function-body
       do (if (or (and (stringp last-elem) (stringp elem))
		     (and (null last-elem) (stringp elem))
		     (subtypep (type-of elem) 'htcomponent)
		     (and (evenp (length attributes)) (stringp elem))
		     body)           
		 (push elem body)		 
		 (push elem attributes)))	     
    (list (reverse attributes) (reverse body))))

(defun generate-id (id)
  "This function is very useful when having references to components id inside component body.
When used with :STATIC-ID the generated id will be mantained as is, and rendered just like the :ID tag attribute." 
  (let* ((id-ht (request-id-table-map))
	 (client-id-index (gethash id id-ht 0))
	 (result))
    (if (= 0 client-id-index)
	(setf result id)
	(setf result (format nil "~a~d" id client-id-index)))
    (setf (gethash id id-ht) (1+ client-id-index))
    result))

(defun build-tagf (tag-name parent empty-p &rest rest)
  "This function is used to create a tag object instance
- TAG-NAME the a string tag name to create, for example \"span\" 
- PARENT the parent class. usually 'TAG
- EMPTY-P determines if the tag must be rendered as an empty tag during the request cycle phase.
- REST a list of attribute/value pairs and the component body"
  (let* ((fbody (parse-htcomponent-function (flatten rest)))
	 (id-table-map (request-id-table-map))
	 (id (getf (first fbody) :id))
	 (static-id (getf (first fbody) :static-id))
	 (instance))
    (unless (null static-id)
      (remf (first fbody) :id)
      (setf id nil))
    (setf instance (make-instance parent 
				  :empty empty-p
				  :name (string-downcase tag-name)
				  :attributes (first fbody)
				  :body (second fbody)))
    (if (null static-id)
	(unless (or (null id-table-map) (null id))
	    (setf (htcomponent-client-id instance)
		  (generate-id id)))
	(setf (htcomponent-client-id instance) static-id))
    instance))

(defun generate-tagf (tag-name empty-p)
  "Internal function that generates an htcomponent creation function from the component class name
- TAG-NAME the symbol class name of the component
- EMPTY-P determines if the tag must be rendered as an empty tag during the request cycle phase."
  (setf (fdefinition (intern (format nil "~a>" (string-upcase tag-name))))
	#'(lambda (&rest rest) (build-tagf tag-name 'tag empty-p rest))))

(defun empty-string-p (str)
  "Determines if a string is empty"
  (string= str ""))

;;;----------------------------------------------------------------



(defclass page()
  ((writer :initarg :writer
	   :accessor page-writer :documentation "The output stream for this page instance")    
   (lisplet :initarg :lisplet
	    :reader page-lisplet :documentation "The lisplet that owns this page instance")
   (can-print :initform nil
	      :accessor page-can-print)
   (script-files :initarg :script-files
		 :accessor page-script-files :documentation "Holds component class scripts files injected by components during the request cycle")
   (stylesheet-files :initarg :stylesheet-files
		     :accessor page-stylesheet-files :documentation "Holds component class  css files injected by components during the request cycle")
   (class-initscripts :initarg :class-initscripts
		      :accessor page-class-initscripts :documentation "Holds component class javascript directives injected by components during the request cycle")
   (instancee-initscripts :initarg :instance-initscripts
			 :accessor page-instance-initscripts :documentation "Holds component instance javascript directives injected by components during the request cycle")  
   (indent :initarg :indent
	   :accessor page-indent :documentation "Determine if the output must be indented or not")
   (tabulator :initarg :tabulator
	   :accessor page-tabulator :documentation "Holds the indentation level")
   (xmloutput :initarg :xmloutput
	  :accessor page-xmloutput :documentation "Determine if the page must be rendered as an XML")
   (current-form :initform :nil
		 :accessor page-current-form :documentation "During the rewinding phase the form or the action-link whose action has been fired")
   (content-type :initarg :doc-type
		 :accessor page-doc-type :documentation "The DOCUMENT TYPE of the page (default to HTML 4.01 STRICT)")
   (lasttag :initform nil 
	     :accessor page-lasttag :documentation "Last rendered tag. Needed for page output rendering")
   (json-component-count :initarg :json-component-count
			 :accessor page-json-component-count :documentation "Need to render the json object after an xhr call.")
   (request-parameters :initarg :request-parameters)
   (url :initarg :url
	:accessor page-url :documentation "The URL provided with this page instance"))
  (:default-initargs :writer t
    :script-files nil 
    :json-component-count 0
    :stylesheet-files nil
    :class-initscripts nil
    :instance-initscripts nil
    :indent t
    :tabulator 0
    :xmloutput nil
    :doc-type *html-4.01-strict*
    :request-parameters nil
    :url nil)
  (:documentation "A page object holds cl-webobjects components to be rendered") )
  
(defclass htcomponent ()
  ;class for html tags
  ((page :initarg :page
	 :reader htcomponent-page :documentation "The owner page") 
   (body :initarg :body
	 :accessor htcomponent-body :documentation "The tag body")
   (client-id :initarg :client-id
	      :accessor htcomponent-client-id :documentation "The tag computed id if :ID war provided for the building function")
   (attributes :initarg :attributes
	       :accessor htcomponent-attributes :documentation "The tag attributes")
   (empty :initarg :empty
	  :accessor htcomponent-empty :documentation "Determine if the tag has to be rendered as an empty tag")
   (script-files :initarg :script-files
		 :accessor htcomponent-script-files :documentation "Page injectable script files")
   (stylesheet-files :initarg :stylesheet-files
		     :accessor htcomponent-stylesheet-files :documentation "Page injectable css files")
   (class-initscripts :initarg :class-initscripts
		     :accessor htcomponent-class-initscripts :documentation "Page injectable javascript class derectives")
   (instance-initscript :initarg :instance-initscript
			:accessor htcomponent-instance-initscript :documentation "Page injectable javascript instance derectives"))
  (:default-initargs :page nil    
    :body nil
    :client-id nil
    :attributes nil 
    :empty nil
    :script-files nil
    :stylesheet-files nil
    :class-initscripts nil
    :instance-initscript nil)
  (:documentation "Base class for all other cl-webobjects components"))

(defclass tag (htcomponent)
  ((name :initarg :name
	 :reader tag-name :documentation "The tag name to be rendered"))
  (:default-initargs :name nil)
  (:documentation "This class is used to render the most part of html tags"))

(defclass htstring (htcomponent)
  ((raw :initarg :raw
	:accessor htstring-raw :documentation "Determines if the string content must be html escaped or not"))
  (:default-initargs :raw nil)
  (:documentation "Component needed to render strings"))

(defmethod initialize-instance :after ((inst tag) &rest keys)
  (let ((empty-p (getf keys :empty))
	(body (getf keys :body)))
    (when (and (not (null empty-p))
	       (not (null body)))
      (error (format nil "This tag cannot have a body <~a> body: '~a'" (tag-name inst) body)))))

(defun $> (value)
  "Creates an escaping htstring component"
  (make-instance 'htstring :body value))

(defun $raw> (value)
  "Creates a non escaping htstring component"
  (make-instance 'htstring :body value :raw t))

(defclass htscript (tag) ()
  (:documentation "Creates a component for rendering a <script> tag"))

(defun script> (&rest rest)
  (build-tagf "script" 'htscript nil rest))
  
(defclass htlink (tag) ()
  (:documentation "Creates a component for rendering a <link> tag"))

(defun link> (&rest rest)
  (build-tagf "link" 'htlink t rest))

(defclass htbody (tag) ()
  (:documentation "Creates a component for rendering a <body> tag"))

(defun body> (&rest rest)
  (build-tagf "body" 'htbody nil rest))

(defclass hthead (tag) ()
  (:documentation "Creates a component for rendering a <head> tag"))

(defun head> (&rest rest)
  (build-tagf "head" 'hthead nil rest))

(mapcar #'(lambda (tag-name) (generate-tagf tag-name t))
	;;Creates empty tag initialization functions. But the ones directly defined
	'("area" "base" "basefont" "br" "col" "frame"
	       "hr" "img" "input" "isindex" "meta"
	       "param"))

(mapcar #'(lambda (tag-name) (generate-tagf tag-name nil))
	;;Creates non empty tag initialization functions. But the ones directly defined
	'("a" "abbr" "acronym" "address" "applet" 
	       "b" "bdo" "big" "blockquote" "button" 
	       "caption" "center" "cite" "code" "colgroup" 
	       "dd" "del" "dfn" "dir" "div" "dl" "dt" 
	       "em" 
	       "fieldset" "font" "form" "frameset"
	       "h1" "h2" "h3" "h4" "h5" "h6" "html"
	       "i" "iframe" "ins"
	       "kbd"
	       "label" "legend" "li"
	       "map" "menu"
	       "noframes" "noscript" 
	       "object" "ol" "optgroup" "option"
	       "p" "pre" 
	       "q" 
	       "s" "samp" "select" "small" "span" "strike" "strong" "style" "sub" "sup"
	       "table" "tbody" "td" "textarea" "tfoot" "th" "thead" "title" "tr" "tt"
	       "u" "ul" "var"))

(defun tag-empty-p (tag-name)
  "Returns if a tag defined by the string TAG-NAME is empty"
  (member tag-name *empty-tags* :test #'string-equal))

(defun tag-symbol-class (tag-name)
  "Returns the symbol class for a given TAG-NAME"
  (let ((name (string-downcase tag-name))) 
    (cond ((string= name "script") 'htscript)
	  ((string= name "link") 'htlink)
	  ((string= name "body") 'htbody)
	  ((string= name "head") 'hthead)
	  (t 'tag))))
;;;--------------------METHODS implementation----------------------------------------------

(defmethod (setf htcomponent-page) ((pobj page) (obj htcomponent))  
  (let ((id (getf (htcomponent-attributes obj) :id))
	(static-id (getf (htcomponent-attributes obj) :static-id)))
    (setf (slot-value obj 'page) pobj)
    (unless (and (null id) (null static-id))
      (let ((client-id (htcomponent-client-id obj)))
	(when (null client-id)
	  (if (null static-id)
	      (setf (htcomponent-client-id obj) (generate-id id))
	      (setf (htcomponent-client-id obj) static-id)))))))

(defmethod page-request-parameters ((pobj page))
  (if (and (boundp '*request*) (null (slot-value pobj 'request-parameters)))
    (let ((parameters (append (post-parameters) (get-parameters)))
	  (pparameters (make-hash-table :test 'equal)))
      (loop for kv in parameters
	   do (setf (gethash (string-upcase (car kv)) pparameters)
		    (append (gethash (string-upcase (car kv)) pparameters) 
			    (list (cdr kv)))))
      (setf (slot-value pobj 'request-parameters) pparameters))
    (slot-value pobj 'request-parameters)))

(defmethod page-req-parameter ((pobj page) name &optional as-list)
  (let ((parameters (page-request-parameters pobj))
	(retval))
    (unless (null parameters)
      (setf retval (gethash (string-upcase name) parameters))
      (if (or (null retval) as-list)
	  retval
	  (first retval)))))

(defmethod page-format ((obj page) str &rest rest)
  (let ((json-p (page-json-id-list obj))
	(writer (page-writer obj)))
    (if (null json-p)
	(apply #'format writer str rest)
	(apply #'format writer (list 
				(regex-replace-all "\""
						   (regex-replace-all "\\\\\""
								      (regex-replace-all "\\n"
											 (apply #'format nil str rest)
											 "\\n")
								      "\\\\\\\"")
						   "\\\""))))))

(defmethod page-format-raw ((obj page) str &rest rest)
  (let ((writer (page-writer obj)))
    (apply #'format writer str rest)))

(defmethod page-json-id-list ((obj page))
  (page-req-parameter obj "json" t))

(defmethod page-init ((obj page))
  (progn
    (reset-request-id-table-map) 
    (setf (page-can-print obj) (null (page-json-id-list obj)))
    (reset-request-id-table-map)
    (setf (page-tabulator obj) 0)))

(defmethod page-render-headings ((obj page))
  (let* ((writer (page-writer obj))
	 (json-p (page-json-id-list obj))
	 (encoding (handler-case (format nil "~a" (stream-external-format writer))
		     (error () (format nil "~a" *default-encoding*))))
	 (xml-p (page-xmloutput obj))
	 (content-type (page-doc-type obj)))    
    (when (null json-p)
      (unless (null xml-p)	  
	(page-format-raw obj "<?xml version=\"1.0\" encoding=\"~a\"?>~%" encoding))	 
      (unless (null content-type)
	(page-format-raw obj "~a~%" content-type)))))
  
(defmethod page-render ((obj page))  
  (let ((body (page-content obj))
	(json-p (page-json-id-list obj)))
    (if (null body)
	(format nil "null body for page ~a~%" (type-of obj))
	(progn
	  (page-init obj)
	  (unless (null (page-req-parameter obj *rewind-parameter*))
	    (htcomponent-rewind body obj))
	  (page-init obj)
	  (htcomponent-prerender (page-content obj) obj) ;Here we need a fresh new body!!!
	  (page-render-headings obj)
	  (page-init obj)	  
	  (unless (null json-p)
	    (page-format-raw obj "{components:{"))
	  (htcomponent-render (page-content obj) obj) ;Here we need a fresh new body!!!
	  (unless (null json-p)
	    (page-format-raw obj "},classInjections:\"")
	    (setf (page-can-print obj) t)	    
	    (dolist (injection (page-init-injections obj))
	      (htcomponent-render injection obj))
	    (page-format-raw obj "\",instanceInjections:\"")
	    (htcomponent-render (htbody-init-scripts-tag obj) obj)
	    (page-format-raw obj "\"}"))))))

(defmethod page-body-init-scripts ((pobj page))
  (let ((js-body ""))
    (dolist (current-js (reverse (page-instance-initscripts pobj)))
      (setf js-body (format nil "~a~%~a~%" js-body current-js)))
    (if (empty-string-p js-body)
	js-body
	(format nil "~a" js-body))))

(defmethod page-print-tabulation ((obj page))
  (let ((json-p (page-json-id-list obj))
	(tabulator (page-tabulator obj))
	(indent-p (page-indent obj)))
    (when (and (<= 0 tabulator) indent-p (null json-p))
      (page-format-raw obj "~a" 
	      (make-string tabulator :initial-element #\tab)))))

(defmethod page-newline ((obj page))
  (let ((json-p (page-json-id-list obj))
	(indent-p (page-indent obj)))
    (when (and indent-p (null json-p))
      (page-format-raw obj "~%")))) 

(defmethod page-init-injections ((pobj page))
    (let ((tag-list)
	  (class-init-scripts ""))      

      (dolist (script (reverse (page-class-initscripts pobj)))
	(setf class-init-scripts (format nil "~a~%~a" 
					 class-init-scripts
					 script)))
      (unless (empty-string-p class-init-scripts)
	(let ((current-js (script> :type "text/javascript")))
	  (setf (htcomponent-body current-js) class-init-scripts)
	  (push current-js tag-list)))
            
      (dolist (js-file (page-script-files pobj))
	(let ((current-js (script> :type "text/javascript" :src "")))
	  (setf (getf (htcomponent-attributes current-js) :src) js-file)
	  (push current-js tag-list)))

      (dolist (css-file (page-stylesheet-files pobj))
	(let ((current-css (link> :rel "stylesheet" :type "text/css" :href "")))
	  (setf (getf (htcomponent-attributes current-css) :href) css-file)
	  (push current-css tag-list)))

      tag-list))

;;;========= HTCOMPONENT ============================
(defmethod htcomponent-can-print ((obj htcomponent))
  (let* ((id (htcomponent-client-id obj))
	 (pobj (htcomponent-page obj))	 
	 (print-status (page-can-print pobj))
	 (render-p (member id (page-json-id-list pobj) :test #'string=)))
    (or print-status render-p)))

(defmethod htcomponent-json-print-start-component ((obj htcomponent))
  (let* ((pobj (htcomponent-page obj))
	 (json-p (page-json-id-list pobj))
	 (id (htcomponent-client-id obj)))
    (unless (or (null json-p) (null (member id json-p :test #'string-equal)))
      (when (> (page-json-component-count pobj) 0)
	(page-format pobj ","))
      (page-format-raw pobj "~a:\"" id)
      (incf (page-json-component-count pobj)))))

(defmethod htcomponent-json-print-end-component ((obj htcomponent))
  (let* ((pobj (htcomponent-page obj))
	 (json-p (page-json-id-list pobj))
	 (id (htcomponent-client-id obj)))
    (unless (or (null json-p) (null (member id json-p :test #'string-equal)))
      (page-format-raw pobj "\""))))

(defmethod htcomponent-rewind :before ((obj htcomponent) (pobj page))
  (setf (htcomponent-page obj) pobj))
(defmethod htcomponent-prerender :before ((obj htcomponent) (pobj page))
  (setf (htcomponent-page obj) pobj))
(defmethod htcomponent-render :before ((obj htcomponent) (pobj page))
  (setf (htcomponent-page obj) pobj))      

(defmethod htcomponent-rewind ((obj htcomponent) (pobj page))
  (dolist (tag (htcomponent-body obj))
    (when (subtypep (type-of tag) 'htcomponent)
      (htcomponent-rewind tag pobj))))

(defmethod htcomponent-prerender ((obj htcomponent) (pobj page))
  (let ((previous-print-status (page-can-print pobj)))        
    (when (null previous-print-status)
      (setf (page-can-print pobj) (htcomponent-can-print obj)))
    (dolist (tag (htcomponent-body obj))
      (when (subtypep (type-of tag) 'htcomponent)
	(htcomponent-prerender tag pobj)))
    (when (null previous-print-status)
      (setf (page-can-print pobj) nil))))

(defmethod htcomponent-render ((obj htcomponent) (pobj page))  
  (let ((body-list (htcomponent-body obj))
	(previous-print-status (page-can-print pobj)))
    (when (null previous-print-status)
      (setf (page-can-print pobj) (htcomponent-can-print obj))
      (htcomponent-json-print-start-component obj))
    (dolist (tag body-list)
      (if (stringp tag)
	  (htcomponent-render ($> tag) pobj)
	  (htcomponent-render tag pobj)))
    (when (null previous-print-status)
      (setf (page-can-print pobj) nil)
      (htcomponent-json-print-end-component obj))))

;;;========= TAG =====================================
(defmethod tag-render-attributes ((obj tag) (pobj page))
  (unless (null (htcomponent-attributes obj))
    (loop for (k v) on (htcomponent-attributes obj) by #'cddr 
       do (progn
	    (assert (keywordp k)) 
	    (unless (null v)		   
	      (page-format pobj " ~a=\"~a\"" 
			   (string-downcase (if (eq k :static-id)
						"id"
						(symbol-name k)))
			   (let ((s (if (eq k :id)
					(prin1-to-string (htcomponent-client-id obj))
					(prin1-to-string v)))) ;escapes double quotes
			     (subseq s 1 (1- (length s))))))))))

(defmethod tag-render-starttag ((obj tag) (pobj page))
  (let ((tagname (tag-name obj))
	(empty-p (htcomponent-empty obj))
	(xml-p (page-xmloutput pobj)))
    (setf (page-lasttag pobj) tagname)
    (page-newline pobj)
    (page-print-tabulation pobj)    
    (page-format pobj "<~a" tagname)    
    (tag-render-attributes obj pobj)
    (if (null empty-p)
	(progn
	  (page-format pobj ">")
	  (incf (page-tabulator pobj)))
	(if (null xml-p)
	    (page-format pobj ">")
	    (page-format pobj "/>")))))
    
(defmethod tag-render-endtag ((obj tag) (pobj page))
  (let ((tagname (tag-name obj))
	 (previous-tagname (page-lasttag pobj))
	 (empty-p (htcomponent-empty obj)))
    (when (null empty-p)
      (progn	
	(decf (page-tabulator pobj))
	(if (string= tagname previous-tagname)
	    (progn
	      (page-format pobj "</~a>" tagname))
	    (progn	      
	      (page-newline pobj)
	      (page-print-tabulation pobj)
	      (page-format pobj "</~a>" tagname)))))
    (setf (page-lasttag pobj) nil)))

(defmethod htcomponent-render ((obj tag) (pobj page))   
  (let ((body-list (htcomponent-body obj))
	(previous-print-status (page-can-print pobj)))
    (when (null previous-print-status)
      (setf (page-can-print pobj) (htcomponent-can-print obj))
      (htcomponent-json-print-start-component obj))
    (unless (or (null (page-can-print pobj)) (null previous-print-status))
      (tag-render-starttag obj pobj))
    (dolist (tag body-list)
      (if (stringp tag)
	    (htcomponent-render ($> tag) pobj)
	    (htcomponent-render tag pobj)))
    (unless (or (null (page-can-print pobj)) (null previous-print-status))
      (tag-render-endtag obj pobj))
    (when (null previous-print-status)
      (setf (page-can-print pobj) nil)
      (htcomponent-json-print-end-component obj))))

;;;========= HTHEAD ======================================
(defmethod htcomponent-render ((obj hthead) (pobj page))
  (when (null (page-json-id-list pobj))          
    (let ((body-list (htcomponent-body obj))
	  (injections (page-init-injections pobj)))
      (tag-render-starttag obj pobj)
      (dolist (tag body-list)	  
	(if (stringp tag)
	    (htcomponent-render ($> tag) pobj)
	    (htcomponent-render tag pobj)))
      (dolist (injection injections)
	(htcomponent-render injection pobj))
      (tag-render-endtag obj pobj))))
  
;;;========= HTSTRING ===================================

(defmethod htcomponent-rewind((obj htstring) (pobj page)))
(defmethod htcomponent-prerender((obj htstring) (pobj page)))

(defmethod htcomponent-render ((obj htstring) (pobj page))
  (let ((body (htcomponent-body obj))
	(json-p (not (null (page-json-id-list pobj))))
	(print-p (page-can-print pobj)))
    (unless (or (null print-p) (null body))
      (unless (null json-p)
	(setf body (regex-replace-all "\""
				      (regex-replace-all "\\\\\""
							 (regex-replace-all "\\n"
									    body
									    "\\n")				  
				      "\\\\\\\"")
				      "\\\"")))
      (if (null (htstring-raw obj))      
	  (loop for ch across body
	     do (case ch
		  ((#\<) (page-format-raw pobj "&#60;"))
		  ((#\>) (page-format-raw pobj "&#62;"))
		  ((#\&) (page-format-raw pobj "&#38;"))		  
		  (t (page-format-raw pobj "~a" ch))))
	  (page-format-raw pobj body)))))

;;;========= HTSCRIPT ===================================
(defmethod htcomponent-prerender((obj htscript) (pobj page)))

(defmethod htcomponent-render ((obj htscript) (pobj page))
  (let ((xml-p (page-xmloutput pobj))
	(body (htcomponent-body obj))
	(previous-print-status (page-can-print pobj)))
    (when (null previous-print-status)
      (setf (page-can-print pobj) (htcomponent-can-print obj))
      (htcomponent-json-print-start-component obj))
    (unless (getf (htcomponent-attributes obj) :type)
      (append '(:type "text/javascript") (htcomponent-attributes obj)))
    (unless (null (page-can-print pobj))
      (tag-render-starttag obj pobj)    
      (when (and (null (getf (htcomponent-attributes obj) :src)) 
		 (not (null (htcomponent-body obj))))
	(if (null xml-p)
	    (page-format pobj "~%//<!--~%")   
	    (page-format pobj "~%//<[CDATA[~%"))
	(unless (listp body)
	  (setf body (list body)))
	(dolist (element body)
	  (if (stringp element)
	      (htcomponent-render ($raw> element) pobj)
	      (htcomponent-render element pobj)))
	(if (null xml-p)
	    (page-format pobj "~%//-->")
	    (page-format pobj "~%//]]>")))
      (setf (page-lasttag pobj) nil)
      (tag-render-endtag obj pobj))
    (when (null previous-print-status)
      (setf (page-can-print pobj) nil)
      (htcomponent-json-print-end-component obj))))

;;;========= HTLINK ====================================

(defmethod htcomponent-render ((obj htlink) (pobj page))
  (let ((previous-print-status (page-can-print pobj)))
    (when (null previous-print-status)
      (setf (page-can-print pobj) (htcomponent-can-print obj))
      (htcomponent-json-print-start-component obj))
    (unless (null (page-can-print pobj))
      (unless (getf (htcomponent-attributes obj) :type)
	(append '(:type "text/css") (htcomponent-attributes obj)))
      (unless (getf (htcomponent-attributes obj) :rel)
	(append '(:rel "styleshhet") (htcomponent-attributes obj)))  
      (tag-render-starttag obj pobj)        
      (tag-render-endtag obj pobj))
    (when (null previous-print-status)
      (setf (page-can-print pobj) nil)
      (htcomponent-json-print-end-component obj))))

;;;========= HTBODY ===================================
(defmethod htcomponent-render ((obj htbody) (pobj page))
  (let ((body-list (htcomponent-body obj))
	(previous-print-status (page-can-print pobj)))    
    (unless (or (null (page-can-print pobj)) (null previous-print-status))
      (setf (page-can-print pobj) (htcomponent-can-print obj))
      (htcomponent-json-print-start-component obj))
    (unless (null (page-can-print pobj))
      (tag-render-starttag obj pobj))
    (dolist (tag body-list)
      (if (stringp tag)
	    (htcomponent-render ($> tag) pobj)
	    (htcomponent-render tag pobj)))
    (unless (null (page-can-print pobj))
      (htcomponent-render (htbody-init-scripts-tag pobj) pobj)
      (tag-render-endtag obj pobj))
    (unless (or (null (page-can-print pobj)) (null previous-print-status))
      (setf (page-can-print pobj) nil)
      (htcomponent-json-print-end-component obj))))
	
(defmethod htbody-init-scripts-tag ((pobj page))
  (let ((js (script> :type "text/javascript")))
    (setf (htcomponent-page js) pobj)    
    (setf (htcomponent-body js) (page-body-init-scripts pobj))
    js))

;;;========= WOCOMPONENT ===================================
(defclass wocomponent (htcomponent)
  ((parameters :initarg :parameters
	       :accessor wocomponent-parameters
	       :type cons
	       :documentation "must be a plist or nil")
   (reserved-parameters :initarg :reserved-parameters
			:accessor wocomponent-reserved-parameters
			:type cons :documentation "Parameters that may not be used in the constructor function")
   (informal-parameters :initarg :informal-parameters
			:accessor wocomponent-informal-parameters
			:type cons :documentation "Informal parameters are parameters optional for the component")
   (allow-informal-parameters :initarg :allow-informal-parameters
			      :reader wocomponent-allow-informal-parametersp
			      :allocation :class :documentation "Determines if the component accepts informal parameters")
   (template :initform nil
	     :accessor wocomponent-template
	     :type htcomponent :documentation "The component template. What gives to each wocomponent its unique aspect and features"))
  (:default-initargs :informal-parameters nil 
    :reserved-parameters nil
    :parameters nil
    :allow-informal-parameters t)
  (:documentation "Base class for creationg customized web components. Use this or one of its subclasses to make your own."))

(defmethod wocomponent-check-parameters((comp wocomponent))
  (let ((id nil)
	(static-id nil))
    (loop for (k v) on (htcomponent-attributes comp) by #'cddr
       do (progn (when (and (eql v ':required) (not (eq k :id)))
		   (error (format nil 
				  "Parameter ~a of class ~a is required" 
				  k (class-name (class-of comp)))))
		 (when (eq k :id)
		   (setf id v))
		 (when (eq k :static-id)
		   (setf static-id v))))
    (when (and (eq id :required) (null static-id))
      (error (format nil 
		     "Parameter id of class ~a is required" 
		     (class-name (class-of comp)))))))
	   

(defun make-component (name parameters content)
  (let ((instance (make-instance name))
	(static-id (getf parameters :static-id)))
    (unless (null static-id)
      (remf parameters :id))
    (loop for (k v) on parameters by #'cddr
       do (let ((keyword k))
	    (when (eq keyword :static-id)
	      (setf keyword :id))
	    (multiple-value-bind (inst-k inst-v inst-p) 
		(get-properties (wocomponent-parameters instance) (list keyword))
	      (declare (ignore inst-v))
	      (unless (null (find inst-k (wocomponent-reserved-parameters instance)))
		(error (format nil "Parameter ~a is reserved" inst-k)))
	      (if (null inst-p)
		  (if (null (wocomponent-allow-informal-parametersp instance))
		      (error (format nil 
				     "Component ~a doesn't accept informal parameters" 
				     name))
		      (setf (getf (wocomponent-informal-parameters instance) keyword) v))
		  (progn
		    (when (and (eq keyword :id) (not (null static-id)))
		      (setf keyword :static-id))
		    (setf (getf (htcomponent-attributes instance) keyword) v))))))
    (wocomponent-check-parameters instance)    
    (let ((id (wocomponent-parameter-value instance :id))
	  (static-id (wocomponent-parameter-value instance :static-id)))
      (if (and (null static-id) id)	  
	  (setf (htcomponent-client-id instance) (generate-id id))
	  (setf (htcomponent-client-id instance) static-id)))
    (setf (htcomponent-body instance) content)
    instance))

(defun build-component (component-name &rest rest)
  (let ((fbody (parse-htcomponent-function (flatten rest))))
    (make-component component-name (first fbody) (second fbody))))


(defmethod wocomponent-parameter-value ((c wocomponent) key)
  (let ((result (getf (htcomponent-attributes c) key :undefined)))
    (if (eq result :undefined)
	(getf (wocomponent-parameters c) key)
	result)))

(defmacro defcomponent (name superclass-name slot-specifier &body class-option)
  (let ((symbolf (intern (format nil "~a>" name))))
    `(eval-when (:compile-toplevel :load-toplevel :execute) 
      (defclass ,name 
	  ,@(if (null superclass-name)
		(list '(wocomponent))
		(list 
		 (let ((result))
		   (dolist (parent superclass-name) 
		     (when (subtypep parent 'wocomponent)
		       (setf result t)))
		   (if result
		       superclass-name
		       (append '(wocomponent) superclass-name)))))
	,@(if (null class-option)
	      (list slot-specifier)
	      (push slot-specifier class-option)))
      (setf (fdefinition `,',symbolf) #'(lambda(&rest rest) (build-component ',name rest))))))
	

(defmethod htcomponent-rewind ((obj wocomponent) (pobj page))
  (let ((template (wocomponent-template obj)))
    (wocomponent-before-rewind obj pobj)
    (if (listp template)
	(dolist (tag template)	    
	  (htcomponent-rewind tag pobj))	  
	(htcomponent-rewind template pobj))
    (wocomponent-after-rewind obj pobj)))

(defmethod wocomponent-before-rewind ((obj wocomponent) (pobj page)))
(defmethod wocomponent-after-rewind ((obj wocomponent) (pobj page)))

(defmethod htcomponent-prerender ((obj wocomponent) (pobj page))
  (wocomponent-before-prerender obj pobj)
  (let ((previous-print-status (page-can-print pobj))
	(template (wocomponent-template obj)))          
    (when (null previous-print-status)
      (setf (page-can-print pobj) (htcomponent-can-print obj)))
    (unless (null (page-can-print pobj))
      (dolist (script (htcomponent-script-files obj))
	(pushnew script (page-script-files pobj) :test #'equal))
      (dolist (css (htcomponent-stylesheet-files obj))
	(pushnew css (page-stylesheet-files pobj) :test #'equal))
      (dolist (js (htcomponent-class-initscripts obj))
	(pushnew js (page-class-initscripts pobj) :test #'equal)) 
      (unless (null (htcomponent-instance-initscript obj))
	(pushnew (htcomponent-instance-initscript obj) (page-instance-initscripts pobj) :test #'equal)))
    (if (listp template)
	(dolist (tag template)	  	  
	  (when (subtypep (type-of tag) 'htcomponent)
	    (htcomponent-prerender tag pobj)))
	(htcomponent-prerender template pobj))
    (when (null previous-print-status)
      (setf (page-can-print pobj) nil)))
  (wocomponent-after-prerender obj pobj))

(defmethod wocomponent-before-prerender ((obj wocomponent) (pobj page)))
(defmethod wocomponent-after-prerender ((obj wocomponent) (pobj page)))

(defmethod htcomponent-render ((obj wocomponent) (pobj page))
  (let ((template (wocomponent-template obj))
	(previous-print-status (page-can-print pobj)))        
    (when (null previous-print-status)
      (setf (page-can-print pobj) (htcomponent-can-print obj))
      (htcomponent-json-print-start-component obj))
    (wocomponent-before-render obj pobj)
    (unless (listp template)
      (setf template (list template)))
    (dolist (tag template)	
      (if (stringp tag)
	  (htcomponent-render ($> tag) pobj)
	  (htcomponent-render tag pobj)))
    (wocomponent-after-render obj pobj)
    (when (null previous-print-status)
      (setf (page-can-print pobj) nil)
      (htcomponent-json-print-end-component obj))))

(defmethod wocomponent-before-render ((obj wocomponent) (pobj page)))
(defmethod wocomponent-after-render ((obj wocomponent) (pobj page)))
