;;; -*- lisp -*-

(in-package :wo-test)


(setf *woserver-base-path* "/clwo")

(defvar *test-lisplet*)
(setf *test-lisplet* (make-instance 'lisplet :realm "test1" :base-path "/test"))

(defvar *test-lisplet2*)
(setf *test-lisplet2* (make-instance 'lisplet :realm "test2" :base-path "/test2"))



(defparameter *woserver* (make-instance 'woserver :port 4242))
;;;(defparameter *woserver* (make-instance 'woserver :port 4242 :sslport 4445 
;;;					:ssl-certificate-file #P"/home/kiuma/pem/cacert.pem" 
;;;					:ssl-privatekey-file #P"/home/kiuma/pem/privkey.pem"))

(woserver-register-lisplet *woserver* *test-lisplet*)
(woserver-register-lisplet *woserver* *test-lisplet2*)

(defun startme ()
  (woserver-start *woserver*))

(defun stopme ()
  (woserver-stop *woserver*))


;;;--------------------template--------------------------------

(defcomponent site-template () ()) 

(defmethod wocomponent-parameters ((o site-template))
  (list :title :required))

(defmethod wocomponent-template ((o site-template))
  (html> 
   (head>
    (title> 
     (wocomponent-parameter-value o ':title)))
   (body>      
    (wocomponent-informal-parameters o)
    (p>
     (a> :href "/clwo/test/index.html"))
    (htcomponent-body o))))


;;;--------------------index testing page--------------------------------
(defclass index-page (page) ())

(defmethod page-content ((o index-page))  
  (site-template> :title "Home test page"
		  (p> :id "p"
		      (ul>
		       (li> (a> :href "realm.html" :target "clwo1" 
				"realm on lisplet 'test'"))
		       (li> (a> :href "../test2/realm.html" :target "clwo2" 
				"realm on lisplet 'test2'"))
		       (li> (a> :href "id-tests.html" "id generation test"))
		       (li> (a> :href "form.html" ($> "form components test")))))))
		   
(lisplet-register-page-location *test-lisplet* 'index-page "index.html")

;;;--------------------realm test page--------------------------------
(defclass realm-page (page) ())

(defmethod page-content ((o realm-page))  
  (site-template> :title "Realm test page"		  
		  (progn
		    (when (null *session*)		      
		      (start-session (format nil "~@[~a~]~a/" *woserver-base-path* (lisplet-base-path (page-lisplet o))))
		      (setf (session-value 'RND-NUMBER) (random 1000)))
		    (p>
		     "session"
		     (ul>
		      (li> "Rnd number value: " (format nil "~d" (session-value 'RND-NUMBER)))
		      (li> "Remote Addr: " (session-remote-addr  *session*))
		      (li> "User agent: " (session-user-agent *session*))
		      (li> "Lisplet Realm: " (lisplet-realm (page-lisplet o)))
		      (li> "Session Realm: " (session-realm *session*))
		      (li> "Session value: " (format nil "~a" (hunchentoot::session-string *session*)))
		      (li> "Request Realm: " (request-realm *request*)))))))
		   
(lisplet-register-page-location *test-lisplet* 'realm-page "realm.html")
(lisplet-register-page-location *test-lisplet2* 'realm-page "realm.html")

;;;--------------------id testing page--------------------------------
(defclass id-tests-page (page) ())

(defmethod page-content ((o id-tests-page))
  (let ((uid (generate-id "uid"))
	(uid2 (generate-id "uid")))
    (site-template> :title "a page title" 
		    "\"<escaping>test\""
		    (hr>)
		    (div> :id "foo" :class "goo" 
			  :onclick "this.innerHTML = this.id"
			  "passed id: 'foo'[click me, to see generated id]")
		    (div> :id "foo" 
			  :onclick "this.innerHTML = this.id"
			  "passed id: 'foo'[click me, to see generated id]")
		    (div> :static-id uid 
			  :onclick "this.innerHTML = this.id"
			  "passed id: 'uid' (generated with generate-id)[click me, to see generated id]")
		    (div> :static-id uid2 
			  :onclick "this.innerHTML = this.id"
			  "passed id: 'uid' (generated with generate-id)[click me, to see generated id]"))))

(lisplet-register-page-location *test-lisplet* 'id-tests-page "id-tests.html")


;;;--------------------from components testing page--------------------------------
(defclass form-page (page) 
  ((name :initarg :name
	 :accessor form-page-name)
   (surname :initarg :surname
	    :accessor form-page-surname)
   (gender :initarg :gender	   
	   :reader form-page-gender
	   :writer setf-gender)
   (colors :initarg :colors
	    :accessor form-page-colors))
  
  (:default-initargs :name "kiuma"
    :surname "surnk"
    :colors nil
    :gender '("M")))

(defmethod page-content ((o form-page))
  (site-template> :title "a page title" 
		  (cform> :id "testform" :method "post"
			  (table>
			   (tr>
			    (td> "Name")
			    (td>
			     (cinput> :id "name"
				      :type "text"
				      :accessor 'form-page-name)))
			   (tr>
			    (td> "Surname")
			    (td>
			     (cinput> :id "surname"
				      :type "text"
				      :accessor 'form-page-surname)))
			   (tr>
			    (td> "Gender")
			    (td>
			     (cselect> :id "gender"				     
				      :writer 'setf-gender
				      (loop for gender in (list "M" "F")
					   collect (option> :value gender
							    (when (string= gender (first (form-page-gender o)))
								'(:selected "selected"))
							    (if (string= gender "M")
								"Male"
								"Female"))))))
			   (tr>
			    (td> "Colors")
			    (td>
			     (cselect> :id "colors"				
				       :multiple "true"
				       :style "width:80px;height:120px;"
				      :accessor 'form-page-colors
				      (loop for color in (list "R" "G" "B")
					   collect (option> :value color
							    (when (member color (form-page-colors o) :test #'string=)
								'(:selected "selected"))
							    (cond 
							      ((string= color "R") "red")
							      ((string= color "G") "green")
							      (t "blue")))))))		
			   (tr>
			    (td> :colspan "2"
				 (csubmit> :id "submit" :value "OK")))))
		  (div> (format nil "Name: ~a" (form-page-name o)))
		  (div> (format nil "Surname: ~a" (form-page-surname o)))
		  (div> (format nil "Gender: ~a" (first (form-page-gender o))))))

(lisplet-register-page-location *test-lisplet* 'form-page "form.html")



